# coding: utf-8

"""
    Anyscale API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from anyscale_client.configuration import Configuration


class StaticSSOConfig(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'idp_entity_id': 'str',
        'idp_sso_url': 'str',
        'idp_sso_binding': 'str',
        'idp_x509cert': 'str'
    }

    attribute_map = {
        'idp_entity_id': 'idp_entity_id',
        'idp_sso_url': 'idp_sso_url',
        'idp_sso_binding': 'idp_sso_binding',
        'idp_x509cert': 'idp_x509cert'
    }

    def __init__(self, idp_entity_id=None, idp_sso_url=None, idp_sso_binding=None, idp_x509cert=None, local_vars_configuration=None):  # noqa: E501
        """StaticSSOConfig - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._idp_entity_id = None
        self._idp_sso_url = None
        self._idp_sso_binding = None
        self._idp_x509cert = None
        self.discriminator = None

        self.idp_entity_id = idp_entity_id
        self.idp_sso_url = idp_sso_url
        self.idp_sso_binding = idp_sso_binding
        self.idp_x509cert = idp_x509cert

    @property
    def idp_entity_id(self):
        """Gets the idp_entity_id of this StaticSSOConfig.  # noqa: E501

        Static configuration for the SAML 2.0 identity provider's entity id.  # noqa: E501

        :return: The idp_entity_id of this StaticSSOConfig.  # noqa: E501
        :rtype: str
        """
        return self._idp_entity_id

    @idp_entity_id.setter
    def idp_entity_id(self, idp_entity_id):
        """Sets the idp_entity_id of this StaticSSOConfig.

        Static configuration for the SAML 2.0 identity provider's entity id.  # noqa: E501

        :param idp_entity_id: The idp_entity_id of this StaticSSOConfig.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and idp_entity_id is None:  # noqa: E501
            raise ValueError("Invalid value for `idp_entity_id`, must not be `None`")  # noqa: E501

        self._idp_entity_id = idp_entity_id

    @property
    def idp_sso_url(self):
        """Gets the idp_sso_url of this StaticSSOConfig.  # noqa: E501

        Static configuration for the SAML 2.0 identity provider's url.  # noqa: E501

        :return: The idp_sso_url of this StaticSSOConfig.  # noqa: E501
        :rtype: str
        """
        return self._idp_sso_url

    @idp_sso_url.setter
    def idp_sso_url(self, idp_sso_url):
        """Sets the idp_sso_url of this StaticSSOConfig.

        Static configuration for the SAML 2.0 identity provider's url.  # noqa: E501

        :param idp_sso_url: The idp_sso_url of this StaticSSOConfig.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and idp_sso_url is None:  # noqa: E501
            raise ValueError("Invalid value for `idp_sso_url`, must not be `None`")  # noqa: E501

        self._idp_sso_url = idp_sso_url

    @property
    def idp_sso_binding(self):
        """Gets the idp_sso_binding of this StaticSSOConfig.  # noqa: E501

        Static configuration for the SAML 2.0 identity provider's sso binding.  # noqa: E501

        :return: The idp_sso_binding of this StaticSSOConfig.  # noqa: E501
        :rtype: str
        """
        return self._idp_sso_binding

    @idp_sso_binding.setter
    def idp_sso_binding(self, idp_sso_binding):
        """Sets the idp_sso_binding of this StaticSSOConfig.

        Static configuration for the SAML 2.0 identity provider's sso binding.  # noqa: E501

        :param idp_sso_binding: The idp_sso_binding of this StaticSSOConfig.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and idp_sso_binding is None:  # noqa: E501
            raise ValueError("Invalid value for `idp_sso_binding`, must not be `None`")  # noqa: E501

        self._idp_sso_binding = idp_sso_binding

    @property
    def idp_x509cert(self):
        """Gets the idp_x509cert of this StaticSSOConfig.  # noqa: E501

        Static configuration for the SAML 2.0 identity provider's x509cert  # noqa: E501

        :return: The idp_x509cert of this StaticSSOConfig.  # noqa: E501
        :rtype: str
        """
        return self._idp_x509cert

    @idp_x509cert.setter
    def idp_x509cert(self, idp_x509cert):
        """Sets the idp_x509cert of this StaticSSOConfig.

        Static configuration for the SAML 2.0 identity provider's x509cert  # noqa: E501

        :param idp_x509cert: The idp_x509cert of this StaticSSOConfig.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and idp_x509cert is None:  # noqa: E501
            raise ValueError("Invalid value for `idp_x509cert`, must not be `None`")  # noqa: E501

        self._idp_x509cert = idp_x509cert

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, StaticSSOConfig):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, StaticSSOConfig):
            return True

        return self.to_dict() != other.to_dict()
